#!D:\bld\gnuradio_1611083683161\_h_env\python.exe
#
# Copyright 2006,2007,2011,2012,2019,2020 Free Software Foundation, Inc.
#
# This file is part of GNU Radio
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
#
"""
This example application demonstrates receiving and demodulating
different types of signals using the USRP.

A receive chain is built up of the following signal processing
blocks:

USRP  - Daughter board source generating complex baseband signal.
CHAN  - Low pass filter to select channel bandwidth
RFSQL - RF squelch zeroing output when input power below threshold
AGC   - Automatic gain control leveling signal at [-1.0, +1.0]
DEMOD - Demodulation block appropriate to selected signal type.
        This converts the complex baseband to real audio frequencies,
        and applies an appropriate low pass decimating filter.
CTCSS - Optional tone squelch zeroing output when tone is not present.
RSAMP - Resampler block to convert audio sample rate to user specified
        sound card output rate.
AUDIO - Audio sink for playing final output to speakers.

The following are required command line parameters:

-f FREQ         USRP receive frequency
-m MOD          Modulation type, select from AM, FM, or WFM

The following are optional command line parameters:

-R SUBDEV       Daughter board specification, defaults to first found
-c FREQ         Calibration offset.  Gets added to receive frequency.
                Defaults to 0.0 Hz.
-g GAIN         Daughterboard gain setting. Defaults to mid-range.
-o RATE         Sound card output rate. Defaults to 32000. Useful if
                your sound card only accepts particular sample rates.
-r RFSQL        RF squelch in db. Defaults to -50.0.
-p FREQ         CTCSS frequency.  Opens squelch when tone is present.

Once the program is running, ctrl-break (Ctrl-C) stops operation.

Please see fm_demod.py and am_demod.py for details of the demodulation
blocks.
"""

import math
import sys
from argparse import ArgumentParser
from gnuradio import gr, uhd, audio
from gnuradio import filter
from gnuradio import analog
from gnuradio import eng_notation, eng_arg
from gnuradio.eng_option import eng_option


# (device_rate, channel_rate, audio_rate, channel_pass, channel_stop, demod)
DEMOD_PARAMS = {
                'AM'  : (256000,  16000, 16000,  5000,   8000, analog.demod_10k0a3e_cf),
                'FM'  : (256000,  32000,  8000,  8000,   9000, analog.demod_20k0f3e_cf),
                'WFM' : (320000, 320000, 32000, 80000, 115000, analog.demod_200kf3e_cf)
               }

class uhd_src(gr.hier_block2):
    """
    Create a UHD source object supplying complex floats.

    Selects user supplied subdevice or chooses first available one.

    Calibration value is the offset from the tuned frequency to
    the actual frequency.
    """
    def __init__(self, args, spec, antenna, samp_rate, gain=None, calibration=0.0):
        gr.hier_block2.__init__(
            self, "uhd_src",
            gr.io_signature(0, 0, 0),                    # Input signature
            gr.io_signature(1, 1, gr.sizeof_gr_complex)) # Output signature

        self._src = \
            uhd.usrp_source(device_addr=args, stream_args=uhd.stream_args('fc32'))

        # Set the subdevice spec
        if spec:
            self._src.set_subdev_spec(spec, 0)

        # Set the antenna
        if antenna:
            self._src.set_antenna(antenna, 0)

        self._src.set_samp_rate(samp_rate)
        dev_rate = self._src.get_samp_rate()
        self._samp_rate = samp_rate

        # Resampler to get to exactly samp_rate no matter what dev_rate is
        self._rrate = samp_rate / dev_rate
        self._resamp = filter.pfb.arb_resampler_ccf(self._rrate)

        # If no gain specified, set to midrange
        gain_range = self._src.get_gain_range()
        if gain is None:
            gain = (gain_range.start()+gain_range.stop())/2.0
            print("Using gain: ", gain)
        self._src.set_gain(gain)

        self._cal = calibration
        self.connect(self._src, self._resamp, self)

    def tune(self, freq):
        """ Set center frequency """
        self._src.set_center_freq(freq + self._cal, 0)

    def rate(self):
        """ Set sampling rate """
        return self._samp_rate

class app_top_block(gr.top_block):
    def __init__(self, options):
        gr.top_block.__init__(self)
        self.options = options

        (dev_rate, channel_rate, audio_rate, channel_pass, channel_stop, demod) = \
            DEMOD_PARAMS[options.modulation]

        dev = uhd_src(options.args,             # UHD device address
                      options.spec,             # device subdev spec
                      options.antenna,          # device antenna
                      dev_rate,                 # device sample rate
                      options.gain,             # Receiver gain
                      options.calibration)      # Frequency offset
        dev.tune(options.frequency)

        if_rate = dev.rate()
        channel_decim = if_rate // channel_rate
        audio_decim = channel_rate // audio_rate

        chan_taps = filter.optfir.low_pass(1.0,          # Filter gain
                                           if_rate,      # Sample rate
                                           channel_pass, # One sided modulation bandwidth
                                           channel_stop, # One sided channel bandwidth
                                           0.1,          # Passband ripple
                                           60)           # Stopband attenuation

        chan = filter.freq_xlating_fir_filter_ccf(
            channel_decim, # Decimation rate
            chan_taps,     # Filter taps
            0.0,           # Offset frequency
            if_rate)       # Sample rate

        rfsql = analog.pwr_squelch_cc(
            options.rf_squelch,    # Power threshold
            125.0/channel_rate,    # Time constant
            int(channel_rate/20),  # 50ms rise/fall
            False)                 # Zero, not gate output

        agc = analog.agc_cc(1.0/channel_rate,  # Time constant
                            1.0,               # Reference power
                            1.0)               # Gain

        demod = demod(channel_rate, audio_decim)

        # From RF to audio
        #self.connect(dev, chan, rfsql, agc, demod)
        self.connect(dev, chan, demod)

        # Optionally add CTCSS and RSAMP if needed
        tail = demod
        if options.ctcss != None and options.ctcss > 60.0:
            ctcss = analog.ctcss_squelch_ff(audio_rate,    # Sample rate
                                            options.ctcss) # Squelch tone
            self.connect(demod, ctcss)
            tail = ctcss

        if options.output_rate != audio_rate:
            out_gcd = math.gcd(audio_rate, options.output_rate)
            out_interp = options.output_rate // out_gcd
            out_decim = audio_rate // out_gcd
            rsamp = filter.rational_resampler_fff(out_interp, out_decim)
            self.connect(tail, rsamp)
            tail = rsamp

        # Send to audio output device
        audio_sink = audio.sink(int(options.output_rate),
                                options.audio_output)
        self.connect(tail, audio_sink)

        # Direct asynchronous notifications to callback function
        if self.options.show_async_msg:
            self.async_msgq = gr.msg_queue(0)
            self.async_src = uhd.amsg_source("", self.async_msgq)
            self.async_rcv = uhd.msgq_runner(self.async_msgq, self.async_callback)

    def async_callback(self, msg):
        """ Callback for processing async messages """
        md = self.async_src.msg_to_async_metadata_t(msg)
        print("Channel: {} Time: {} Event: {}".format(
            md.channel,
            md.time_spec.get_real_secs(),
            md.event_code))


def main():
    """ Go, Go, Go """
    parser = ArgumentParser()
    parser.add_argument("-a", "--args", default="",
                        help="UHD device address args")
    parser.add_argument("--spec",
                        help="Subdevice of UHD device where appropriate")
    parser.add_argument("-A", "--antenna",
                        help="select Rx Antenna where appropriate")
    parser.add_argument("-f", "--frequency", type=eng_arg.eng_float,
                        default=None, metavar="Hz",
                        help="set receive frequency to Hz")
    parser.add_argument("-c", "--calibration", type=eng_arg.eng_float,
                        default=0.0, metavar="Hz",
                        help="set frequency offset to Hz")
    parser.add_argument("-g", "--gain", type=eng_arg.eng_float,
                        metavar="dB", default=None,
                        help="set RF gain [default is midpoint]")
    parser.add_argument("-m", "--modulation", choices=('AM', 'FM', 'WFM'),
                        metavar="TYPE",
                        help="set modulation type (AM,FM,WFM)")
    parser.add_argument("-o", "--output-rate", type=eng_arg.eng_float,
                        default=32000, metavar="RATE",
                        help="set audio output rate to RATE")
    parser.add_argument("-r", "--rf-squelch", type=eng_arg.eng_float,
                        default=-50.0, metavar="dB",
                        help="set RF squelch to dB [default=%(default)s]")
    parser.add_argument("-p", "--ctcss", type=float,
                        default=None, metavar="FREQ",
                        help="set CTCSS squelch to FREQ")
    parser.add_argument("-O", "--audio-output", default="default",
                        help="pcm device name.  E.g., hw:0,0 or surround51 or /dev/dsp")
    parser.add_argument("--show-async-msg", action="store_true",
                        help="Show asynchronous message notifications from UHD")
    args = parser.parse_args()

    if args.frequency is None:
        sys.stderr.write("Must supply receive frequency with -f.\n")
        sys.exit(1)

    if args.modulation is None:
        sys.stderr.write("Must supply a modulation type (AM, FM, WFM).\n")
        sys.exit(1)

    tb = app_top_block(args)
    try:
        tb.run()
    except KeyboardInterrupt:
        pass

if __name__ == "__main__":
    main()



#!/usr/bin/env python3
#
# Copyright 2015-2016 Free Software Foundation, Inc.
#
# This file is part of GNU Radio
#
# SPDX-License-Identifier: GPL-3.0-or-later
#
#
"""
USRP Helper Module: Common tasks for uhd-based apps.
"""

import sys
import time
import argparse
from gnuradio import eng_arg
from gnuradio import uhd
from gnuradio import gr

COMMAND_DELAY = .2 # Seconds

COMPACT_TPL = "{mb_id} ({mb_serial}), {db_subdev} ({subdev}, {ant}{db_serial})"
LONG_TPL = """{prefix}  Motherboard: {mb_id} ({mb_serial})
{prefix}  Daughterboard: {db_subdev}{db_serial}
{prefix}  Subdev: {subdev}
{prefix}  Antenna: {ant}
"""

# PyLint can't reliably detect C++ exports in modules, so let's disable that
# pylint: disable=no-member

class UHDApp:
    GAIN_TYPE_GAIN = 'dB'
    GAIN_TYPE_POWER = 'power_dbm'

    " Base class for simple UHD-based applications "
    def __init__(self, prefix=None, args=None):
        self.prefix = prefix
        self.args = args
        self.verbose = args.verbose or 0
        if self.args.sync == 'auto' and len(self.args.channels) > 1:
            self.args.sync = 'pps'
        self.antenna = None
        self.gain_range = None # Can also be power range
        self.samp_rate = None
        self.has_lo_sensor = None
        self.async_msgq = None
        self.async_src = None
        self.async_rcv = None
        self.tr = None
        self.gain = None
        self.freq = None
        self.channels = None
        self.cpu_format = None
        self.spec = None
        self.clock_source = None
        self.time_source = None
        self.lo_source = None
        self.lo_export = None
        self.usrp = None
        self.lo_source_channel = None
        self.gain_type = self.GAIN_TYPE_GAIN

    def vprint(self, *args):
        """
        Print 'string' with 'prefix' prepended if self.verbose is True
        """
        if self.verbose:
            print("[{prefix}]".format(prefix=self.prefix), *args)

    def get_usrp_info_string(self, compact=False, tx_or_rx='rx', chan=0, mboard=0):
        """
        Return a nice textual description of the USRP we're using.
        """
        if tx_or_rx not in ['rx', 'tx']:
            raise ValueError("tx_or_rx argument must be one of ['rx', 'tx']")

        try:
            info_pp = {}
            if self.prefix is None:
                info_pp['prefix'] = ""
            else:
                info_pp['prefix'] = "[{prefix}] ".format(prefix=self.prefix)
            usrp_info = self.usrp.get_usrp_info(chan)
            info_pp['mb_id'] = usrp_info['mboard_id']
            info_pp['mb_serial'] = usrp_info['mboard_serial']
            if info_pp['mb_serial'] == "":
                info_pp['mb_serial'] = "no serial"
            info_pp['db_subdev'] = usrp_info["{xx}_subdev_name".format(xx=tx_or_rx)]
            info_pp['db_serial'] = ", " + usrp_info["{xx}_serial".format(xx=tx_or_rx)]
            if info_pp['db_serial'] == "":
                info_pp['db_serial'] = "no serial"
            info_pp['subdev'] = self.usrp.get_subdev_spec(mboard)
            info_pp['ant'] = self.usrp.get_antenna(chan)
            if info_pp['mb_id'] in ("B200", "B210", "E310"):
                # In this case, this is meaningless
                info_pp['db_serial'] = ""
            tpl = LONG_TPL
            if compact:
                tpl = COMPACT_TPL
            return tpl.format(**info_pp)
        except Exception:
            return "Can't establish USRP info."

    def normalize_sel(self, num_name, arg_name, num, arg):
        """
        num_name: meaningful name why we need num arguments
        arg_name: name of current argument
        num: required number of arguments
        arg: actual argument
        """
        if arg is None:
            return None
        args = [x.strip() for x in arg.split(",")]
        if len(args) == 1:
            args = args * num
        if len(args) != num:
            raise ValueError("Invalid {m} setting for {n} {b}: {a}".format(
                m=arg_name, n=num, a=arg, b=num_name
            ))

        return args

    def async_callback(self, msg):
        """
        Call this when USRP async metadata needs printing.
        """
        metadata = self.async_src.msg_to_async_metadata_t(msg)
        print("[{prefix}] Channel: {chan} Time: {t} Event: {e}".format(
            prefix=self.prefix,
            chan=metadata.channel,
            t=metadata.time_spec.get_real_secs(),
            e=metadata.event_code,
        ))

    def setup_usrp(self, ctor, args, cpu_format='fc32'):
        """
        Instantiate a USRP object; takes care of all kinds of corner cases and settings.
        Pop it and some args onto the class that calls this.
        """
        self.channels = args.channels
        self.cpu_format = cpu_format
        # Create a UHD device object:
        self.usrp = ctor(
            device_addr=args.args,
            stream_args=uhd.stream_args(
                cpu_format,
                args.otw_format,
                args=args.stream_args,
                channels=self.channels,
            )
        )
        # Set the subdevice spec:
        self.spec = self.normalize_sel("mboards", "subdev",
                                       self.usrp.get_num_mboards(), args.spec)
        if self.spec:
            for mb_idx in range(self.usrp.get_num_mboards()):
                self.usrp.set_subdev_spec(self.spec[mb_idx], mb_idx)
        # Set the clock and/or time source:
        if args.clock_source is not None:
            self.clock_source = self.normalize_sel("mboards", "clock-source",
                                                   self.usrp.get_num_mboards(), args.clock_source)
            for mb_idx in range(self.usrp.get_num_mboards()):
                self.usrp.set_clock_source(self.clock_source[mb_idx], mb_idx)
        if args.time_source is not None:
            self.time_source = self.normalize_sel("mboards", "time-source",
                                                  self.usrp.get_num_mboards(), args.time_source)
            for mb_idx in range(self.usrp.get_num_mboards()):
                self.usrp.set_time_source(self.time_source[mb_idx], mb_idx)
        # Sampling rate:
        self.usrp.set_samp_rate(args.samp_rate)
        self.samp_rate = self.usrp.get_samp_rate()
        self.vprint("Using sampling rate: {rate}".format(rate=self.samp_rate))
        # Set the antenna:
        self.antenna = self.normalize_sel("channels", "antenna", len(args.channels), args.antenna)
        if self.antenna is not None:
            for i, chan in enumerate(self.channels):
                if not self.antenna[i] in self.usrp.get_antennas(i):
                    print("[ERROR] {} is not a valid antenna name for this USRP device!".format(self.antenna[i]))
                    sys.exit(1)
                self.usrp.set_antenna(self.antenna[i], i)
                self.vprint("[{prefix}] Channel {chan}: Using antenna {ant}.".format(
                    prefix=self.prefix, chan=chan, ant=self.usrp.get_antenna(i)
                ))
        self.antenna = self.usrp.get_antenna(0)
        # Set receive daughterboard gain:
        if args.power:
            if args.gain is not None:
                print("[ERROR] Providing both --gain and --power is invalid!")
                sys.exit(1)
            self.gain_type = self.GAIN_TYPE_POWER
            if not self.usrp.has_power_reference(0):
                print("[ERROR] Device does not have power reference capabilities!")
                sys.exit(1)
            self.set_power_reference(args.power)
            self.gain_range = self.usrp.get_power_range(0)
        else:
            self.set_gain(args.gain)
            self.gain_range = self.usrp.get_gain_range(0)
            self.gain_type = self.GAIN_TYPE_GAIN
        # Set frequency (tune request takes lo_offset):
        if hasattr(args, 'lo_offset') and args.lo_offset is not None:
            treq = uhd.tune_request(args.freq, args.lo_offset)
        else:
            treq = uhd.tune_request(args.freq)
        self.has_lo_sensor = 'lo_locked' in self.usrp.get_sensor_names()
        # Set LO export and LO source operation
        if (args.lo_export is not None) and (args.lo_source is not None):
            self.lo_source = self.normalize_sel("channels", "lo-source", len(self.channels), args.lo_source)
            self.lo_export = self.normalize_sel("channels", "lo-export", len(self.channels), args.lo_export)
            self.lo_source_channel = None
            for chan, lo_source, lo_export in zip(self.channels, self.lo_source, self.lo_export):
                if (lo_source == "None") or (lo_export == "None"):
                    continue
                if lo_export == "True":
                    #If channel is LO source set frequency and store response
                    self.usrp.set_lo_export_enabled(True, uhd.ALL_LOS, chan)
                if lo_source == "internal":
                    self.lo_source_channel = chan
                    tune_resp = self.usrp.set_center_freq(treq, chan)
                self.usrp.set_lo_source(lo_source, uhd.ALL_LOS, chan)
            # Use lo source tune response to tune dsp_freq on remaining channels
            if self.lo_source_channel is not None:
                if getattr(args, 'lo_offset', None) is not None:
                    treq = uhd.tune_request(
                        target_freq=args.freq,
                        rf_freq=args.freq+args.lo_offset,
                        rf_freq_policy=uhd.tune_request.POLICY_MANUAL,
                        dsp_freq=tune_resp.actual_dsp_freq,
                        dsp_freq_policy=uhd.tune_request.POLICY_MANUAL)
                else:
                    treq = uhd.tune_request(
                        target_freq=args.freq,
                        rf_freq=args.freg,
                        rf_freq_policy=uhd.tune_request.POLICY_MANUAL,
                        dsp_freq=tune_resp.actual_dsp_freq,
                        dsp_freq_policy=uhd.tune_request.POLICY_MANUAL)
            for chan in args.channels:
                if chan == self.lo_source_channel:
                    continue
                self.usrp.set_center_freq(treq, chan)
        # Make sure tuning is synched:
        command_time_set = False
        if len(self.channels) > 1:
            if args.sync == 'pps':
                self.usrp.set_time_unknown_pps(uhd.time_spec())
            cmd_time = self.usrp.get_time_now() + uhd.time_spec(COMMAND_DELAY)
            try:
                for mb_idx in range(self.usrp.get_num_mboards()):
                    self.usrp.set_command_time(cmd_time, mb_idx)
                command_time_set = True
            except RuntimeError:
                sys.stderr.write('[{prefix}] [WARNING] Failed to set command times.\n'.format(prefix=self.prefix))
        for i, chan in enumerate(self.channels):
            self.tr = self.usrp.set_center_freq(treq, i)
            if self.tr is None:
                sys.stderr.write(
                    '[{prefix}] [ERROR] Failed to set center frequency on channel {chan}\n'
                    .format(prefix=self.prefix, chan=chan))
                sys.exit(1)
        if command_time_set:
            for mb_idx in range(self.usrp.get_num_mboards()):
                self.usrp.clear_command_time(mb_idx)
            self.vprint("Syncing channels...".format(prefix=self.prefix))
            time.sleep(COMMAND_DELAY)
        self.freq = self.usrp.get_center_freq(0)
        if args.show_async_msg:
            self.async_msgq = gr.msg_queue(0)
            self.async_src = uhd.amsg_source("", self.async_msgq)
            self.async_rcv = uhd.msgq_runner(self.async_msgq, self.async_callback)

    def set_gain(self, gain):
        """
        Safe gain-setter. Catches some special cases:
        - If gain is None, set to mid-point in dB.
        - If the USRP is multi-channel, set it on all channels.
        """
        if gain is None:
            if self.args.verbose:
                self.vprint("Defaulting to mid-point gains:".format(prefix=self.prefix))
            for i, chan in enumerate(self.channels):
                self.usrp.set_normalized_gain(.5, i)
                if self.args.verbose:
                    self.vprint("Channel {chan} gain: {g} dB".format(
                        chan=chan, g=self.usrp.get_gain(i)
                    ))
        else:
            self.vprint("Setting gain to {g:.2f} dB.".format(g=gain))
            for chan in range(len(self.channels)):
                self.usrp.set_gain(gain, chan)
        self.gain = self.usrp.get_gain(0)

    def set_power_reference(self, power_dbm):
        """
        Safe power-ref-setter.
        """
        assert power_dbm is not None
        self.vprint("Setting ref power to {g:.2f} dBm.".format(g=power_dbm))
        self.usrp.set_power_reference(power_dbm)
        self.gain = self.usrp.get_power_reference(0)

    def set_freq(self, freq, skip_sync=False):
        """
        Safely tune all channels to freq.
        """
        self.vprint("Tuning all channels to {freq} MHz.".format(freq=freq / 1e6))
        # Set frequency (tune request takes lo_offset):
        if hasattr(self.args, 'lo_offset') and self.args.lo_offset is not None:
            treq = uhd.tune_request(freq, self.args.lo_offset)
        else:
            treq = uhd.tune_request(freq)
        # Special TwinRX tuning due to LO sharing
        if getattr(self, 'lo_source_channel', None) is not None:
            tune_resp = self.usrp.set_center_freq(treq, self.lo_source_channel)
            if getattr(self.args, 'lo_offset', None) is not None:
                treq = uhd.tune_request(
                    target_freq=freq,
                    rf_freq=freq+self.args.lo_offset,
                    rf_freq_policy=uhd.tune_request.POLICY_MANUAL,
                    dsp_freq=tune_resp.actual_dsp_freq,
                    dsp_freq_policy=uhd.tune_request.POLICY_MANUAL)
            else:
                treq = uhd.tune_request(
                    target_freq=freq,
                    rf_freq=freq,
                    rf_freq_policy=uhd.tune_reqest.POLICY_MANUAL,
                    dsp_freq=tune_resp.actual_dsp_freq,
                    dsp_freq_policy=uhd.tune_request.POLICY_MANUAL)
            for chan in self.channels:
                if chan == self.lo_source_channel:
                    continue
                self.usrp.set_center_freq(treq, chan)
        # Make sure tuning is synched:
        command_time_set = False
        if len(self.channels) > 1 and not skip_sync:
            cmd_time = self.usrp.get_time_now() + uhd.time_spec(COMMAND_DELAY)
            try:
                for mb_idx in range(self.usrp.get_num_mboards()):
                    self.usrp.set_command_time(cmd_time, mb_idx)
                command_time_set = True
            except RuntimeError:
                sys.stderr.write('[{prefix}] [WARNING] Failed to set command times.\n'.format(prefix=self.prefix))
        for i, chan in enumerate(self.channels):
            self.tr = self.usrp.set_center_freq(treq, i)
            if self.tr is None:
                sys.stderr.write('[{prefix}] [ERROR] Failed to set center frequency on channel {chan}\n'.format(
                    prefix=self.prefix, chan=chan
                ))
                sys.exit(1)
        if command_time_set:
            for mb_idx in range(self.usrp.get_num_mboards()):
                self.usrp.clear_command_time(mb_idx)
            self.vprint("Syncing channels...".format(prefix=self.prefix))
            time.sleep(COMMAND_DELAY)
        self.freq = self.usrp.get_center_freq(0)
        self.vprint("First channel has freq: {freq} MHz.".format(freq=self.freq / 1e6))

    @staticmethod
    def setup_argparser(
            parser=None,
            description='USRP App',
            allow_mimo=True,
            tx_or_rx="",
            skip_freq=False,
        ):
        """
        Create or amend an argument parser with typical USRP options.
        """
        def cslist(string):
            """
            For ArgParser: Turn a comma separated list into an actual list.
            """
            try:
                return [int(x.strip()) for x in string.split(",")]
            except ValueError:
                raise argparse.ArgumentTypeError("Not a comma-separated list: {string}".format(string=string))
        if parser is None:
            parser = argparse.ArgumentParser(
                description=description,
            )
        tx_or_rx = tx_or_rx.strip() + " "
        group = parser.add_argument_group('USRP Arguments')
        group.add_argument("-a", "--args", default="", help="UHD device address args")
        group.add_argument(
            "--spec",
            help="Subdevice(s) of UHD device where appropriate. "
                 "Use a comma-separated list to set different boards to different specs.")
        group.add_argument("-A", "--antenna", help="Select {xx}antenna(s) where appropriate".format(xx=tx_or_rx))
        group.add_argument("-s", "--samp-rate", type=eng_arg.eng_float, default=1e6,
                           help="Sample rate")
        group.add_argument("-g", "--gain", type=eng_arg.eng_float, default=None,
                           help="Gain (default is midpoint)")
        group.add_argument("-p", "--power", type=eng_arg.eng_float, default=None,
                           help="(Reference) power level (in dBm). "
                                "Not supported by all devices (see UHD manual). "
                                "Will fail if not supported. Precludes --gain. "
                                "Behaviour may differ between applications.")
        if not skip_freq:
            group.add_argument("-f", "--freq", type=eng_arg.eng_float, default=None, required=True,
                               help="Set carrier frequency to FREQ",
                               metavar="FREQ")
            group.add_argument("--lo-offset", type=eng_arg.eng_float, default=0.0,
                               help="Set daughterboard LO offset to OFFSET [default=hw default]")
        if allow_mimo:
            group.add_argument("-c", "--channels", default=[0,], type=cslist,
                               help="Select {xx} Channels".format(xx=tx_or_rx))
        group.add_argument(
            "--lo-export",
            help="Set TwinRX LO export {None, True, False} for each channel "
                 "with a comma-separated list. None skips a channel.")
        group.add_argument(
            "--lo-source",
            help="Set TwinRX LO source {None, internal, companion, external} "
                 "for each channel with a comma-separated list. None skips this channel.")
        group.add_argument("--otw-format", choices=['sc16', 'sc12', 'sc8'], default='sc16',
                           help="Choose over-the-wire data format")
        group.add_argument("--stream-args", default="", help="Set additional stream arguments")
        group.add_argument("-v", "--verbose", action="count", help="Use verbose console output")
        group.add_argument("--show-async-msg", action="store_true",
                           help="Show asynchronous message notifications from UHD")
        group.add_argument("--sync", choices=('default', 'pps', 'auto'),
                           default='auto', help="Set to 'pps' to sync devices to PPS")
        group.add_argument("--clock-source",
                           help="Set the clock source; typically 'internal', 'external' or 'gpsdo'")
        group.add_argument("--time-source",
                           help="Set the time source")
        return parser
